# covid19_equil
# Simplifed model based on version 2020-04-15 
# simplying: individual are myopic in the sense they only care about discounted utility of being infected next period

# Jiaying Gu
# April 16, 2020
# modified on May 7 to fix the change in flow chart in paper version 2020-05-06

# revision Jan 30, 2021
# fix a big in infec prob when update status for Xpath

# Logit CDF
LamdaCDF = function(x) exp(x)/(1+exp(x))

## 1. Parameters 
numind = 10000   # number of individuals in the soceity
numinfec_0 = 10   # number of individuals infected at the initial period
ncontacts = 40   # number of contacts per individual
mrate = 0.01 	  # mortality rate 
numt = 150  # number of periods 
reprate = 3.5*(1/6+1/7)     # reproduction rate of the virus 

# Health shock 
pi_negA = 1/6
pi_posA = 1/7
pi_posSU = 1/14
pi_negSU = (10/90) * pi_posSU
pi_posSD = 1/10
pi_negSD = (5/95) * pi_posSD

# Policy parameters on testing 
pi_TestA = 0
pi_TestS = 0


set.seed(1121)
R = 10
Q = matrix(0, R,numt)
S = array(0, c(numt, 10,R))
G = matrix(0,R,numt)
for (rr in 1:R){

eta0 = 0
eta1 = -1
## Production function 
# output (individual) = alpha_a + beta0_a * (active team members) + beta1_a *(inactive team members) + r_a * Q
# parameters: alpha_0, alpha_1, beta0_0, beta0_1, beta1_1, beta1_0, r_0, r_1


Full = log(0.005/(1-0.005))/(-0.65)
alpha_1 = 0.05 * Full
alpha_0 = 0.2 * Full
beta0_1 = (0.3/ ncontacts) * Full
beta0_0 = (0.8/ncontacts) * Full
beta1_1 = (0.15/ncontacts) * Full 
beta1_0 =  (0.2/ncontacts) * Full

alpha_tilda = alpha_1 - alpha_0
beta0_tilda = beta0_1 - beta0_0
beta1_tilda = beta1_1 - beta1_0


# subsidy for working from home (same for everyone)
# nosubsidy
tau_sub = 0 * Full



## Preferences towards risk of infections 

delta = 0.99
delta = log(delta/(1-delta)) - alpha_tilda - beta1_tilda * ncontacts


# Possible states
# x = -5 means dead 
# x = -4 means SD (symptomatic, diagn)
# x = -3 means SU (symptomatic, not diagn)
# x = -2 means AD (asymptomatic, diagn) 
# x = -1 means AU (asymptomatic, not diagn)
# x = 0 means H (healthy)
# x = 1 means RAU (immune, recover from AU)
# x = 2 means RAD (immune, recover from AD or AU)
# x = 3 means RSU (immune, recover from SU)
# x = 4 means RSD (immune, recover from SD or from SU)


xstates = c(-5, -4, -3,-2,-1,0,1,2, 3, 4 )
numstates = length(xstates) 

# initial states 
# xvec1 is vector of numind * 1 with the states of all individuals at t = 1 
# we consider that in each team, all individuals are initially healthy 

xvec1 = rep(0, numind) 
I0ind = sample(numind, numinfec_0)  # randomly sample numinfec_0 individuals to be of state AU
xvec1[I0ind] = -1



# compute equilibrium 
# Eq mapping in Q 
# Q = sum_i w_{i} Lambda(\phi_{0i} + phi_{1i} * Q + phi_{2i} * Lambda(phi_{3i} + phi_{4i} * Q) # this is prob of expected number, instead of expectation of probability
# Q = sum_i w_{i} Lambda(\phi_{0i} + phi_{1i} * Q + phi_{2i} * g(ni, Q), ni is the sum of contacts with state {H,AU,RAU}
# a0 = S_t(AU)/(S_t(AU) + S_t(RAU) + S_t(H))


equilmapQ = function(qval, wi, ni, a0, p0, p1, p2, p3, p4){
	p_AU1 = a0 * (1-qval)
	p_infec = rep(NA, numind)
	for (i in 1:numind){
		p_infec[i] = sum(dbinom(0:ni[i], ni[i], p_AU1)* (1- (1-LamdaCDF(p3 + p4))^(0:ni[i])))
		}
	p_conf = p0 + p1 * qval + p2 * p_infec
	p_conf = 1/(1+exp(-p_conf))
	qbest = sum(wi * p_conf)
	list(qbest = qbest, p_conf = p_conf, p_infec = p_infec)
	}
	

# Matrix with the indexes of the set of contacts of each individual i 

matcontacts = list()
# ring network
for (i in 1:numind){
	first_left = i - (ncontacts/2)
	first_right = i+1 
	icont = c(c(first_left: (first_left + ncontacts/2-1)), c(first_right: (first_right + ncontacts/2-1)))
	icont = (icont >= 1) * (icont <= numind) * icont + (icont <= 0) *(icont + numind) + (icont >= (numind + 1)) * (icont - numind)
	matcontacts[[i]] = icont
	}

# Initialize vectors and matrices of X, S, Q

Qpath = rep(0, numt)
Spath = matrix(0, numt, numstates)
xpath = matrix(0, numt, numind)
new_diagnosed = rep(0, numt)
gdp = rep(0, numt)

xpath[1,] = xvec1

for (t in 1:numt){
	# Generating aggregate frequencies Spath[t,] given vector of individual states xpath[t,] 
	Spath[t,] = table(factor(xpath[t,], levels = xstates))/numind
	# Matrix with, for each individual i, the x state of every member of her set of contacts 
	ixmat = list()
	for (i in 1:numind){
		ixmat[[i]] = xpath[t, matcontacts[[i]]]
		}
in_1 = unlist(lapply(ixmat, function(x) sum(x== -1)))  # vector with for each individual i, the number of her contacts in state -1
	
	
	in0 = unlist(lapply(ixmat, function(x) sum(x== 0))) # vector with for each individual i, the number of her contacts in state 0
	
	in1 = unlist(lapply(ixmat, function(x) sum(x== 1))) # vector with for each individual i, the number of her contacts in state 1
	
	
	in2 = unlist(lapply(ixmat, function(x) sum(x== 2))) # vector with for each individual i, the number of her contacts in state 2
	
	
	in3 = unlist(lapply(ixmat, function(x) sum(x== 3))) # vector with for each individual i, the number of her contacts in state 3
	
	
	in4 = unlist(lapply(ixmat, function(x) sum(x== 4))) # vector with for each individual i, the number of her contacts in state 4
	
	
	in_3 = unlist(lapply(ixmat, function(x) sum(x== -3))) # vector with for each individual i, the number of her contacts in state -3
	
	
	in_2 = unlist(lapply(ixmat, function(x) sum(x== -2)))# vector with for each individual i, the number of her contacts in state -2
	
	
	# S values
	s_5 = Spath[t,1]
	s_4 = Spath[t,2]
	s_3 = Spath[t,3]
	s_2 = Spath[t,4]
	s_1 = Spath[t,5]
	s0 = Spath[t,6]
	s1 = Spath[t,7]
	s2 = Spath[t,8]
	s3 = Spath[t,9]
	s4 = Spath[t,10]
	
	# Eq mapping in Q 
	# w_i = (n_1[i] + n0[i] + n1[i]) /(S_1 + S0 + S1)
	# phi_{0i} = alpha + tau + beta0 * (n_1[i] + n0[i] + n1[i]) + beta0 * n^Rall + beta1 * n^AD
	# phi_{1i} = -beta0 * (n_1[i] + n0[i] + n1[i]) + beta1 * ( n_1[i] + n0[i] + n1[i])
	# phi_{2i} = delta * S0 /(S0 + S_1 + S1)
	# phi_{3i} = eta_0 
	# phi_{4i} = eta_1 
	
	weights = (in_1 + in0 + in1)/(sum(in_1 + in0 + in1))  #1{x_it \in {H,AU,RAU}}/sum_i 1(x_it \in {H,AU,RAU})
	in_HAURAU = (in_1 + in0 + in1) # this is ni
	in_Rall = in2+in3+in4
	in_AD = in_2
	a0 = s_1/(s_1 + s0 + s1)
	phi0 = alpha_tilda + tau_sub + beta0_tilda * in_Rall + beta0_tilda * in_HAURAU + beta1_tilda * in_AD
	phi1 = -beta0_tilda *in_HAURAU + beta1_tilda * in_HAURAU
	phi2 = delta * (s0/(s0+s_1+s1) )
	phi3 = eta0 
	phi4 = eta1
	
	# comment out: contraction mapping for fixed point 
	#qs = seq(0,1,length = 20)
	#contract = qs
	#for (j in 1:length(qs)){
	#	p_temp = rep(NA, numind)
	#for (i in 1:numind){
	#	p_temp[i] = sum(dbinom(0:in_HAURAU[i], in_HAURAU[i], a0 * (1-qs[j]))* plogis(phi3 + phi4 * (0:in_HAURAU[i])))
	#	}
	#p_conftemp = phi0 + phi1 * qs[j] + phi2 * p_temp
	#	contract[j] = sum(weights * plogis(p_conftemp))
	#	}
	#plot(qs,contract,type="l")
	#abline(c(0,1))

	# Computing fixed point Q
	q0 = 0.9
	criter = 1000
	cconv = 1e-06
	maxiter = 1000
	for (r in 1:maxiter){
		if (criter <= cconv){
		#	print(paste("Convergence achieved at iteration", r, "with Fixed point Q =", eqiter$qbest, seq = ""))
			break
			}else{
		eqiter = equilmapQ(q0, weights, in_HAURAU, a0, phi0, phi1, phi2, phi3, phi4)
		criter = abs(eqiter$qbest - q0)
		q0 = eqiter$qbest
		}
		}
	if (criter > cconv){
		#print("Convergence not achieved: Program Terminated")
		}
					
	Qpath[t] = eqiter$qbest
	ip_conf = eqiter$p_conf
	ip_infec = eqiter$p_infec
	
	# GDP - aggregate output [uniform random draws for confinement]
	u_confine = runif(numind)
	active = (u_confine > ip_conf)
	active = (xpath[t,] == -1) * active + (xpath[t,] == 0) * active + (xpath[t,] == 1) * active + (xpath[t,] == 2) + (xpath[t,]==3) + (xpath[t,] ==4) 
	in_a0 = ((in_1 + in0 + in1) * (1-Qpath[t]) + in_Rall)
	in_a1 = ((in_1 + in0 + in1) * Qpath[t] + in_2)  # only AD allowed to work among all disgnoised
	iproduct = active * (alpha_0 + beta0_0 * in_a0 + beta1_0 * in_a1) + (1 - active) * (alpha_1 + beta0_1 * in_a0 + beta1_1 * in_a1)
	gdp[t] = sum(iproduct)
	
	# generating state x[t+1] 
	if (t <= (numt-1)){
		
	 u_infec = runif(numind)
	 
	 	 in_1_active = rep(NA, numind)
	 active = (u_confine > ip_conf)
	 
	 ixmat_active = list()
	for (i in 1:numind){
		ixmat_active[[i]] = active[matcontacts[[i]]] * xpath[t, matcontacts[[i]]]
		}
		
	 in_1_active = unlist(lapply(ixmat_active, function(x) sum(x== -1)))
	 ip_infec_real = 1 - (1- LamdaCDF(phi3 + phi4))^(in_1_active)


	 uu1 = runif(numind)  # health shock
	 uu2 = runif(numind)  # test shock
	
	xpath[t+1,] = (xpath[t,]== -5) * (-5) + (xpath[t,] == 3) * 3 + (xpath[t,] ==4) * 4 + (xpath[t,] == 2) * 2 + (xpath[t,] == 1) * 1 + (xpath[t,] == 0) * (u_confine > ip_conf) * (u_infec <= ip_infec_real) * (-1) + (xpath[t,]==0) * (u_confine > ip_conf) * (u_infec <= ip_infec_real) * 0 + (xpath[t,]==0) * (u_confine <= ip_conf) * 0 + (xpath[t,] ==-1) * (uu1 <= pi_negA) * (uu2 <= pi_TestA) * (-4) + (xpath[t,] == -1) * (uu1> pi_negA & uu1 <= pi_negA + pi_posA) * (uu2 <= pi_TestA) * (1) + (xpath[t,]== -1) * (uu1> pi_posA + pi_negA) * (uu2 <= pi_TestA) * (-2) + (xpath[t,]==-1) * (uu1 <= pi_negA) * (uu2 > pi_TestA) * (-3) + (xpath[t,]==-1) * (uu1> pi_negA & uu1<= pi_negA + pi_posA) * (uu2 > pi_TestA) *(1) + (xpath[t,]==-1) * (uu1> pi_posA + pi_negA) * (uu2 > pi_TestA)*(-1)  +(xpath[t,]==-2) * (uu1 <= pi_negA) * (-4) + (xpath[t,]==-2) * (uu1 > pi_negA & uu1 <= pi_negA + pi_posA) * (2) + (xpath[t,] == -2) * (uu1 > pi_negA + pi_posA) * (-2) + (xpath[t,] == -3) * (uu1 <= pi_negSU)*(uu2 <= pi_TestS) * (-5) + (xpath[t,]==-3)*(uu1 > pi_negSU & uu1<= pi_negSU + pi_posSU) * (uu2 <= pi_TestS) * (3) + (xpath[t,] == -3) * (uu1 > pi_negSU + pi_posSU) * (uu2 <= pi_TestS) * (-4) + (xpath[t,]==-3) * (uu1 <= pi_negSU)*(uu2 > pi_TestS) *(-5) + (xpath[t,]==-3) * (uu1 > pi_negSU & uu1<= pi_negSU + pi_posSU) * (uu2 > pi_TestS)*(3) + (xpath[t,]==-3) *  (uu1 > pi_negSU + pi_posSU) * (uu2 > pi_TestS) *(-3)  + (xpath[t,] == -4) * (uu1 <= pi_negSD) * (-5) + (xpath[t,]==-4) * (uu1 > pi_negSD & uu1 <= pi_negSD + pi_posSD) * (4) + (xpath[t,] == -4) * (uu1 > pi_negSD + pi_posSD) * (-4)
	new_diagnosed[t+1] = sum((xpath[t,]== -1) * (uu1> pi_posA + pi_negA) * (uu2 <= pi_TestA)) + sum((xpath[t,] ==-1) * (uu1 <= pi_negA) * (uu2 <= pi_TestA)) + sum((xpath[t,] == -3) * (uu1 > pi_negSU + pi_posSU) * (uu2 <= pi_TestS))
		}
		}    
S[,,rr] = Spath 
Q[rr,] = Qpath		
G[rr,] = gdp
print(rr)
}	

G.avg = colMeans(G)
Q.avg = colMeans(Q)
S.avg = matrix(NA, numt, 10)
	for (t in 1:numt){
		S.avg[t,] = rowMeans(S[t,,])
		}
 	
### plotting 
timetrend = 1:numt
pdf("NN0_aggre.pdf", height = 10, width = 15)
par(mfrow=c(2,4))
plot(timetrend, Q.avg, type="o", cex = 0.5, xlab = "Day", ylab = "Qt", main = "Share of Confined Individuals Q(t)")
axis(side = 1, at = 10*(1:15))
plot(timetrend, S.avg[,1], type="o", cex = 0.5,xlab = "Day",ylab = "S_t(D)", main = "Share of dead individuals")
axis(side = 1, at = 10*(1:15))
plot(timetrend, S.avg[,7]+S.avg[,8] + S.avg[,9] + S.avg[,10] , xlab = "Day",type="o", ylab = "Share of immune", cex = 0.5,  main = "Share of immune individuals ")
axis(side = 1, at = 10*(1:15))
plot(timetrend, cumsum(c(numinfec_0,-diff(S.avg[,6] * numind)))/numind, xlab = "Day",ylab = "share of infected", type="o", cex = 0.5,  main = "Share of individuals infected")
axis(side = 1, at = 10*(1:15))
plot(timetrend, c(numinfec_0,-diff(S.avg[,6] * numind)), type="o", xlab = "Day",ylab = "cases", cex = 0.5,  main = "number of new infection per day")
axis(side = 1, at = 10*(1:15))
plot(timetrend, 100 * G.avg/G.avg[1], type="o", cex = 0.5,  xlab = "Day",main ="Aggregate Output (period1 = 100)", ylab ="GDP")
axis(side = 1, at = 10*(1:15))
plot(timetrend, log(cumsum(c(numinfec_0,-diff(S.avg[,6] * numind)))), xlab = "Day",ylab = "log infected cases", type="o", cex = .5, main = "log of cumu infected")
axis(side = 1, at = 10*(1:15))
plot(timetrend, log(S.avg[,1]*numind), type="o", cex = .5, xlab = "Day",ylab = "cases", main = "log of cumu death")

dev.off()	



